#!/usr/bin/env python3

import getopt
import logging
import os
import re
import sys

from pkg_resources import parse_version
from urllib.parse import parse_qs


def fatal(msg):
    print(msg, file=sys.stderr)
    sys.exit(1)


def usage():
    fatal('''Usage: manifest_edit [options] [--python_pipeline_config=<filename>] <input>

 --python_pipeline_path=<dirname>     manifest edit pipeline config path.
 --python_pipeline_config=<filename>  manifest edit pipeline config file.
 <input>                              .mpd manifest input file or URL
 [-o outfile]                         output file.
 [-v level]                           0=quiet 1=error 2=warning 3=info 4=debug
 [--python_path path]                 manifest edit python modules search path.
                                      If not provided, defaults to the current
                                      system PYTHONPATH plus the current dir.

Please note that the input file specification may need quoting when extra
parameters are appended in URL style.

This program is part of USP (Unified Streaming Platform). For more information
and its license, please visit http://www.unified-streaming.com
''')


def get_format_from_yaml(yaml_name):
    try:
        with open(yaml_name, 'r') as yaml_file:
            for line in yaml_file:
                before, sep, after = line.partition(':')
                if sep == ':' and before in ('mpd', 'm3u8_main', 'm3u8_media'):
                    return before
        fatal(f'ERROR: unknown manifest format in {yaml_name}')
    except Exception:
        fatal(f'ERROR: Can\'t open {yaml_name}')


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger(__name__)

    try:
        import libfmp4
    except Exception:
        fatal(f'ERROR: python extension "libfmp4" not installed, please refer to the installation documentation')

    try:
        import schema
    except Exception:
        fatal(f'ERROR: python module "schema" not installed, please refer to the installation documentation')
    req_version = '0.7.2'
    if parse_version(schema.__version__) < parse_version(req_version):
        fatal(f'ERROR: python module {schema} is too old (installed version: {schema.__version__}, required: {req_version}), please refer to the installation documentation')

    try:
        import yaml
    except Exception:
        fatal(f'ERROR: python module "yaml" not installed, please refer to the installation documentation')
    req_version = '5.3.1'
    if parse_version(yaml.__version__) < parse_version(req_version):
        fatal(f'ERROR: python module {yaml} is too old (installed version: {yaml.__version__}, required: {req_version}), please refer to the installation documentation')

    shortopts = 'ho:v:'
    longopts = [
        'license_key=',
        'license-key=',
        'show_license=',
        'show-license=',
        'help',
        'python_pipeline_config=',
        'python-pipeline-config=',
        'python_pipeline_path=',
        'python-pipeline-path=',
        'python_path=',
        'python-path=',
        'loop_count=',
        'loop-count=',
    ]
    try:
        logger.debug(f'calling getopt on {sys.argv[1:]}')
        opts, args = getopt.getopt(sys.argv[1:], shortopts, longopts)
    except getopt.GetoptError as err:
        print(f'{sys.argv[0]}: {err}', file=sys.stderr)
        usage()

    license_key = os.getenv('UspLicenseKey')
    show_license = False
    python_pipeline_path = None
    python_pipeline_config = None
    python_path = None
    loop_count = 1
    output_file = 'stdout:'
    verbose = 3 # INFO

    for opt, arg in opts:
        opt = opt.replace('_', '-')
        if opt == '--license-key':
            license_key = arg
        elif opt == '--show-license':
            show_license = True
        elif opt in ('-h', '--help'):
            usage()
        elif opt == '--python-pipeline-path':
            python_pipeline_path = arg
        elif opt == '--python-pipeline-config':
            python_pipeline_config = arg
        elif opt == '--python-path':
            python_path = arg
        elif opt == '--loop-count':
            loop_count = arg
        elif opt == '-o':
            output_file = arg
        elif opt == '-v':
            try:
                verbose = int(arg)
            except ValueError:
                verbose = -1
            # 0=ERROR 1=WARNING 2=NOTICE 3=INFO 4=DEBUG 5=TRACE
            if verbose < 0 or verbose > 5:
                print(f'ERROR: invalid verbose level "{arg}"', file=sys.stderr)
                usage()
        else:
            usage()

    if verbose >= 3:
        print(f'manifest_edit {libfmp4.version_string()}    Copyright 2007-2021 CodeShop B.V.\n', file=sys.stderr)

    if verbose == 0:
        logger.setLevel(logging.ERROR)
    elif verbose == 1:
        logger.setLevel(logging.WARNING)
    elif verbose == 2 or verbose == 3:
        logger.setLevel(logging.INFO)
    elif verbose > 4:
        logger.setLevel(logging.DEBUG)

    if True:
        logger.debug(f'license_key={license_key}')
        logger.debug(f'show_license={show_license}')
        logger.debug(f'python_pipeline_path={python_pipeline_path}')
        logger.debug(f'python_pipeline_config={python_pipeline_config}')
        logger.debug(f'python_path={python_path}')
        logger.debug(f'loop_count={loop_count}')
        logger.debug(f'output_file={output_file}')
        logger.debug(f'verbose={verbose}')
        logger.debug(f'args={args}')

    if len(args) == 0:
        print('ERROR: no input filename provided!', file=sys.stderr)
        usage()

    if len(args) > 1:
        print('ERROR: too many input filenames provided!', file=sys.stderr)
        usage()

    if not python_pipeline_config:
        # If no python_pipeline_config was specified on the command line, look in the environment
        # for QUERY_STRING. If this is set, and has a query argument python_pipeline_config=basename
        # use that as the basename.
        if 'QUERY_STRING' in os.environ:
            query_dict = parse_qs(os.environ['QUERY_STRING'])
            if 'python_pipeline_config' in query_dict:
                query_arg = query_dict['python_pipeline_config'][0]
                if query_arg:
                    # check the argument's sanity. It must have one or more parts separated by
                    # slashes, each part consisting only of numbers, letters, dashes or underscores.
                    match = re.fullmatch(r'([0-9A-Z_a-z-]+)(/([0-9A-Z_a-z-]+))*', query_arg)
                    if not match:
                        fatal(f'ERROR: invalid python_pipeline_config query argument: {query_arg}')
                    python_pipeline_config = match.group(0) + '.yaml'

    if not python_pipeline_config:
        print('ERROR: no valid python pipeline configuration file provided!', file=sys.stderr)
        usage()

    if python_pipeline_path:
        python_pipeline_config = os.path.join(python_pipeline_path, python_pipeline_config)

    logger.debug(f'get format from {python_pipeline_config}')
    format = get_format_from_yaml(python_pipeline_config)
    input_file = args[0]

    if True:
        logger.debug(f'format={format}')
        logger.debug(f'input_file={input_file}')

    if python_path is not None:
        logger.debug(f'inserting {python_path} into sys.path')
        sys.path.insert(1, python_path)

    logger.debug(f'loading manifest_edit entrypoints')
    import manifest_edit.entrypoints as ep
    logger.debug(f'setting up entrypoints with {python_pipeline_config}')
    ep.setup(None, python_pipeline_config)

    if format == 'mpd':
        logger.debug(f'loading mpd manifest {input_file}')
        fmp4_manifest = libfmp4.mpd.load_manifest(input_file)
    elif format == 'm3u8_main':
        logger.debug(f'loading hls master playlist {input_file}')
        fmp4_manifest = libfmp4.hls.load_master_playlist(input_file)
    else:
        logger.debug(f'loading hls media playlist {input_file}')
        fmp4_manifest = libfmp4.hls.load_media_playlist(input_file)

    logger.debug(f'updating fmp4_manifest={fmp4_manifest}')
    new_manifest = ep.update_manifest(fmp4_manifest)

    logger.debug(f'updated new_manifest={new_manifest}')

    if format == 'mpd':
        logger.debug(f'saving mpd manifest {output_file}')
        libfmp4.mpd.save_manifest(new_manifest, output_file)
    elif format == 'm3u8_main':
        logger.debug(f'saving hls master playlist {output_file}')
        libfmp4.hls.save_master_playlist(new_manifest, output_file)
    else:
        logger.debug(f'saving hls media playlist {output_file}')
        libfmp4.hls.save_media_playlist(new_manifest, output_file)

    logger.debug('done')
